from inspect import getmembers
from pprint import pprint
from types import FunctionType

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scvelo as scv
import seaborn as sns
from anndata._core.anndata import AnnData
from beartype import beartype
from scvi.data import synthetic_iid
from termcolor import colored

# import torch
# from scipy.sparse import issparse
# from sklearn.decomposition import PCA
# from torch.nn.functional import relu
# from pyrovelocity.models._transcription_dynamics import inv
# from pyrovelocity.models import mrna_dynamics


def mae(pred_counts, true_counts):
    """
    Computes the mean average error between predicted counts and true counts.

    Args:
        pred_counts (np.ndarray): Predicted counts.
        true_counts (np.ndarray): True counts.

    Returns:
        float: Mean average error.

    Examples:
        >>> import numpy as np
        >>> pred_counts = np.array([[1, 2], [3, 4]])
        >>> true_counts = np.array([[2, 3], [4, 5]])
        >>> mae(pred_counts, true_counts)
        1.0
    """
    error = np.abs(true_counts - pred_counts).sum()
    total = pred_counts.shape[0] * pred_counts.shape[1]
    return (error / total).mean().item()


def ensure_numpy_array(obj):
    return obj.toarray() if hasattr(obj, "toarray") else obj


def mae_evaluate(posterior_samples, adata):
    maes_list = []
    labels = []
    if isinstance(posterior_samples, tuple):
        for model_label, model_obj, split_index in zip(
            ["Poisson train", "Poisson valid"],
            posterior_samples[:2],
            posterior_samples[2:],
        ):
            for sample in range(model_obj["u"].shape[0]):
                maes_list.append(
                    mae(
                        np.hstack(
                            [model_obj["u"][sample], model_obj["s"][sample]]
                        ),
                        np.hstack(
                            [
                                adata.layers["raw_unspliced"].toarray()[
                                    split_index
                                ],
                                adata.layers["raw_spliced"].toarray()[
                                    split_index
                                ],
                            ]
                        ),
                    )
                )
                labels.append(model_label)
    else:
        for model_label, model_obj in zip(
            ["Poisson all cells"], [posterior_samples]
        ):
            for sample in range(model_obj["u"].shape[0]):
                maes_list.append(
                    mae(
                        np.hstack(
                            [model_obj["u"][sample], model_obj["s"][sample]]
                        ),
                        np.hstack(
                            [
                                ensure_numpy_array(
                                    adata.layers["raw_unspliced"]
                                ),
                                ensure_numpy_array(adata.layers["raw_spliced"]),
                            ]
                        ),
                    )
                )
                labels.append(model_label)

    import pandas as pd

    fig, ax = plt.subplots()
    fig.set_size_inches(10, 4)
    df = pd.DataFrame({"MAE": maes_list, "label": labels})
    sns.boxplot(x="label", y="MAE", data=df, ax=ax)
    ax.tick_params(axis="x", rotation=90)
    print(df.groupby("label").mean())
    return df


def attributes(obj):
    """
    get object attributes
    """
    disallowed_names = {
        name
        for name, value in getmembers(type(obj))
        if isinstance(value, FunctionType)
    }
    return {
        name: getattr(obj, name)
        for name in dir(obj)
        if name[0] != "_"
        and name not in disallowed_names
        and hasattr(obj, name)
    }


def print_attributes(obj):
    """
    print object attributes
    """
    pprint(attributes(obj))


@beartype
def pretty_log_dict(d: dict) -> str:
    dict_as_string = ""
    for key, value in d.items():
        key_colored = colored(key, "green")
        value_lines = str(value).split("\n")
        value_colored = "\n".join(
            colored(line, "white") for line in value_lines
        )
        dict_as_string += f"{key_colored}:\n{value_colored}\n"
    return dict_as_string


@beartype
def pretty_print_dict(d: dict):
    print(pretty_log_dict(d))


def filter_startswith_dict(dictionary_with_underscore_keys):
    """Remove entries from a dictionary whose keys start with an underscore.

    Args:
        dictionary_with_underscore_keys (dict): Dictionary to be filtered.

    Returns:
        dict: Filtered dictionary.
    """
    return {
        k: v
        for k, v in dictionary_with_underscore_keys.items()
        if not k.startswith("_")
    }


def print_anndata(anndata_obj):
    """
    Print a formatted representation of an AnnData object.

    This function produces a custom output for the AnnData object with each
    element of obs, var, uns, obsm, varm, layers, obsp, varp indented and
    displayed on a new line.

    Args:
        anndata_obj (AnnData): The AnnData object to be printed.

    Raises:
        AssertionError: If the input object is not an instance of AnnData.

    Examples:
        >>> import anndata
        >>> import numpy as np
        >>> import pandas as pd
        >>> np.random.seed(42)
        >>> X = np.random.randn(10, 5)
        >>> obs = pd.DataFrame({"clusters_coarse": np.random.randint(0, 2, 10),
        ...                     "clusters": np.random.randint(0, 2, 10),
        ...                     "S_score": np.random.rand(10),
        ...                     "G2M_score": np.random.rand(10)})
        >>> var = pd.DataFrame({"gene_name": [f"gene_{i}" for i in range(5)]})
        >>> adata = AnnData(X, obs=obs, var=var)
        >>> print_anndata(adata)  # doctest: +NORMALIZE_WHITESPACE
        AnnData object with n_obs × n_vars = 10 × 5
            obs:
                clusters_coarse,
                clusters,
                S_score,
                G2M_score,
            var:
                gene_name,
    """
    assert isinstance(
        anndata_obj, AnnData
    ), "Input object must be of type AnnData."

    def format_elements(elements):
        formatted = "\n".join([f"        {elem}," for elem in elements])
        return formatted

    print(
        f"AnnData object with n_obs × n_vars = {anndata_obj.n_obs} × {anndata_obj.n_vars}"
    )

    properties = {
        "obs": anndata_obj.obs.columns,
        "var": anndata_obj.var.columns,
        "uns": anndata_obj.uns.keys(),
        "obsm": anndata_obj.obsm.keys(),
        "varm": anndata_obj.varm.keys(),
        "layers": anndata_obj.layers.keys(),
        "obsp": anndata_obj.obsp.keys(),
        "varp": anndata_obj.varp.keys(),
    }

    for prop_name, elements in properties.items():
        if len(elements) > 0:
            print(f"    {prop_name}:\n{format_elements(elements)}")


def generate_sample_data(
    n_obs: int = 100,
    n_vars: int = 12,
    alpha: float = 5,
    beta: float = 0.5,
    gamma: float = 0.3,
    alpha_: float = 0,
    noise_model: str = "gillespie",
    random_seed: int = 0,
) -> AnnData:
    """
    Generate synthetic single-cell RNA sequencing data with spliced and unspliced layers.
    If using the "iid" noise model, the data will be generated with scvi.data.synthetic_iid.
    If using the "normal" or "gillespie" noise model, the data will be generated with
    scvelo.datasets.simulation accounting for the given expression dynamics parameters.

    Args:
        n_obs (int, optional): Number of observations (cells). Default is 100.
        n_vars (int, optional): Number of variables (genes). Default is 12.
        alpha (float, optional): Transcription rate. Default is 5.
        beta (float, optional): Splicing rate. Default is 0.5.
        gamma (float, optional): Degradation rate. Default is 0.3.
        alpha_ (float, optional): Additional transcription rate. Default is 0.
        noise_model (str, optional): Noise model to be used. Must be one of 'iid', 'gillespie', or 'normal'. Default is 'gillespie'.
        random_seed (int, optional): Random seed for reproducibility. Default is 0.

    Returns:
        AnnData: An AnnData object containing the generated synthetic data.

    Raises:
        ValueError: If noise_model is not one of 'iid', 'gillespie', or 'normal'.

    Examples:
        >>> from pyrovelocity.utils import generate_sample_data, print_anndata
        >>> adata = generate_sample_data(random_seed=99)
        >>> print_anndata(adata)
        >>> adata = generate_sample_data(n_obs=50, n_vars=10, noise_model="normal", random_seed=99)
        >>> print_anndata(adata)
        >>> adata = generate_sample_data(n_obs=50, n_vars=10, noise_model="iid", random_seed=99)
        >>> print_anndata(adata)
        >>> adata = generate_sample_data(noise_model="wishful thinking")
        Traceback (most recent call last):
            ...
        ValueError: noise_model must be one of 'iid', 'gillespie', 'normal'
    """
    if noise_model == "iid":
        adata = synthetic_iid(
            batch_size=n_obs,
            n_genes=n_vars,
            n_batches=1,
            n_labels=1,
        )
        adata.layers["spliced"] = adata.X.copy()
        adata.layers["unspliced"] = adata.X.copy()
    elif noise_model in {"gillespie", "normal"}:
        adata = scv.datasets.simulation(
            random_seed=random_seed,
            n_obs=n_obs,
            n_vars=n_vars,
            alpha=alpha,
            beta=beta,
            gamma=gamma,
            alpha_=alpha_,
            noise_model=noise_model,
        )
    else:
        raise ValueError(
            "noise_model must be one of 'iid', 'gillespie', 'normal'"
        )
    return adata


def anndata_counts_to_df(adata):
    spliced_df = pd.DataFrame(
        ensure_numpy_array(adata.layers["raw_spliced"]),
        index=list(adata.obs_names),
        columns=list(adata.var_names),
    )
    unspliced_df = pd.DataFrame(
        ensure_numpy_array(adata.layers["raw_unspliced"]),
        index=list(adata.obs_names),
        columns=list(adata.var_names),
    )

    spliced_melted = spliced_df.reset_index().melt(
        id_vars="index", var_name="var_name", value_name="spliced"
    )
    unspliced_melted = unspliced_df.reset_index().melt(
        id_vars="index", var_name="var_name", value_name="unspliced"
    )

    df = spliced_melted.merge(unspliced_melted, on=["index", "var_name"])

    df = df.rename(columns={"index": "obs_name"})

    total_obs = adata.n_obs
    total_var = adata.n_vars

    max_spliced = adata.layers["raw_spliced"].max()
    max_unspliced = adata.layers["raw_unspliced"].max()

    return (
        df,
        total_obs,
        total_var,
        max_spliced,
        max_unspliced,
    )


# TODO: remove unused functions
# def log(x):
#     """
#     Computes the element-wise natural logarithm of a tensor, while clipping
#     its values to avoid numerical instability.

#     Args:
#         x (torch.Tensor): Input tensor.

#     Returns:
#         torch.Tensor: Tensor with element-wise natural logarithm of x.

#     Examples:
#         >>> import torch
#         >>> x = torch.tensor([0.0001, 0.5, 0.9999])
#         >>> log(x)
#         tensor([-9.2103e+00, -6.9315e-01, -1.0002e-04])
#     """
#     eps = torch.finfo(x.dtype).eps
#     return torch.log(x.clamp(eps, 1 - eps))


# TODO: remove unused functions
# def tau_inv(
#     u=None, s=None, u0=None, s0=None, alpha=None, beta=None, gamma=None
# ):
#     """
#     Computes the inverse tau given the parameters and initial conditions.

#     Args:
#         u (torch.Tensor): Value of u.
#         s (torch.Tensor): Value of s.
#         u0 (torch.Tensor): Initial value of u.
#         s0 (torch.Tensor): Initial value of s.
#         alpha (torch.Tensor): Alpha parameter.
#         beta (torch.Tensor): Beta parameter.
#         gamma (torch.Tensor): Gamma parameter.

#     Returns:
#         torch.Tensor: Inverse tau.

#     Examples:
#         >>> import torch
#         >>> u = torch.tensor(0.6703)
#         >>> s = torch.tensor(0.4596)
#         >>> u0 = torch.tensor(1.0)
#         >>> s0 = torch.tensor(0.5)
#         >>> alpha = torch.tensor(0.5)
#         >>> beta = torch.tensor(0.4)
#         >>> gamma = torch.tensor(0.3)
#         >>> tau_inv(u, s, u0, s0, alpha, beta, gamma)
#         tensor(3.9736e-07)
#     """
#     beta_ = beta * inv(gamma - beta)
#     xinf = alpha / gamma - beta_ * (alpha / beta)
#     tau1 = (
#         -1.0 / gamma * log((s - beta_ * u - xinf) * inv(s0 - beta_ * u0 - xinf))
#     )

#     uinf = alpha / beta
#     tau2 = -1 / beta * log((u - uinf) * inv(u0 - uinf))
#     tau = torch.where(beta > gamma, tau1, tau2)
#     return relu(tau)


# TODO: remove unused functions
# def init_with_all_cells(
#     adata,
#     input_type="knn",
#     shared_time=True,
#     latent_factor="linear",
#     latent_factor_size=10,
#     plate_size=2,
#     num_aux_cells=200,
#     init_smooth=True,
# ):
#     ## hard to use unsmoothed data for initialization
#     ## always initialize the model with smoothed data
#     if "Mu" in adata.layers and "Ms" in adata.layers and init_smooth:
#         u_obs = torch.tensor(adata.layers["Mu"], dtype=torch.float32)
#         s_obs = torch.tensor(adata.layers["Ms"], dtype=torch.float32)
#     elif "spliced" in adata.layers:
#         u_obs = torch.tensor(
#             adata.layers["unspliced"].toarray()
#             if issparse(adata.layers["unspliced"])
#             else adata.layers["unspliced"],
#             dtype=torch.float32,
#         )
#         s_obs = torch.tensor(
#             adata.layers["spliced"].toarray()
#             if issparse(adata.layers["spliced"])
#             else adata.layers["spliced"],
#             dtype=torch.float32,
#         )
#     else:
#         raise

#     ub_u = torch.stack(
#         [torch.quantile(u[u > 0], 0.99) for u in torch.unbind(u_obs, axis=1)]
#     )
#     ub_s = torch.stack(
#         [torch.quantile(s[s > 0], 0.99) for s in torch.unbind(s_obs, axis=1)]
#     )
#     s_mask = (s_obs > 0) & (s_obs <= ub_s)
#     u_mask = (u_obs > 0) & (u_obs <= ub_u)
#     # include zeros
#     training_mask = s_mask & u_mask

#     u_scale = torch.stack(
#         [u[u > 0].std() for u in torch.unbind(u_obs * training_mask, axis=1)]
#     )
#     s_scale = torch.stack(
#         [s[s > 0].std() for s in torch.unbind(s_obs * training_mask, axis=1)]
#     )
#     scale = u_scale / s_scale

#     lb_steady_u = torch.stack(
#         [
#             torch.quantile(u[u > 0], 0.98)
#             for u in torch.unbind(u_obs * training_mask, axis=1)
#         ]
#     )
#     lb_steady_s = torch.stack(
#         [
#             torch.quantile(s[s > 0], 0.98)
#             for s in torch.unbind(s_obs * training_mask, axis=1)
#         ]
#     )

#     steady_u_mask = training_mask & (u_obs >= lb_steady_u)
#     steady_s_mask = training_mask & (s_obs >= lb_steady_s)

#     u_obs = u_obs / scale
#     u_inf = (u_obs * (steady_u_mask | steady_s_mask)).sum(dim=0) / (
#         steady_u_mask | steady_s_mask
#     ).sum(dim=0)
#     s_inf = (s_obs * steady_s_mask).sum(dim=0) / steady_s_mask.sum(dim=0)

#     gamma = (u_obs * steady_s_mask * s_obs).sum(axis=0) / (
#         (steady_s_mask * s_obs) ** 2
#     ).sum(axis=0) + 1e-6
#     gamma = torch.where(gamma < 0.05 / scale, gamma * 1.2, gamma)
#     gamma = torch.where(gamma > 1.5 / scale, gamma / 1.2, gamma)
#     alpha = gamma * s_inf
#     beta = alpha / u_inf

#     switching = tau_inv(u_inf, s_inf, 0.0, 0.0, alpha, beta, gamma)
#     tau = tau_inv(u_obs, s_obs, 0.0, 0.0, alpha, beta, gamma)
#     tau = torch.where(tau >= switching, switching, tau)
#     tau_ = tau_inv(u_obs, s_obs, u_inf, s_inf, 0.0, beta, gamma)
#     tau_ = torch.where(
#         tau_ >= tau_[s_obs > 0].max(dim=0)[0],
#         tau_[s_obs > 0].max(dim=0)[0],
#         tau_,
#     )
#     ut, st = mrna_dynamics(tau, 0.0, 0.0, alpha, beta, gamma)
#     ut_, st_ = mrna_dynamics(tau_, u_inf, s_inf, 0.0, beta, gamma)

#     u_scale_ = u_scale / scale
#     state_on = ((ut - u_obs) / u_scale_) ** 2 + ((st - s_obs) / s_scale) ** 2
#     state_off = ((ut_ - u_obs) / u_scale_) ** 2 + ((st_ - s_obs) / s_scale) ** 2
#     cell_gene_state_logits = state_on - state_off
#     cell_gene_state = cell_gene_state_logits < 0
#     t = torch.where(cell_gene_state_logits < 0, tau, tau_ + switching)
#     init_values = {
#         "alpha": alpha,
#         "beta": beta,
#         "gamma": gamma,
#         "switching": switching,
#         "latent_time": t,
#         "u_scale": u_scale,
#         "s_scale": s_scale,
#         "u_inf": u_inf,
#         "s_inf": s_inf,
#     }
#     if input_type == "knn":
#         init_values["mask"] = training_mask

#     init_values["cell_gene_state"] = cell_gene_state.int()
#     if latent_factor == "linear":
#         if "spliced" in adata.layers:
#             u_obs = torch.tensor(
#                 adata.layers["unspliced"].toarray()
#                 if issparse(adata.layers["unspliced"])
#                 else adata.layers["unspliced"],
#                 dtype=torch.float32,
#             )
#             s_obs = torch.tensor(
#                 adata.layers["spliced"].toarray()
#                 if issparse(adata.layers["spliced"])
#                 else adata.layers["spliced"],
#                 dtype=torch.float32,
#             )
#         test = np.hstack([u_obs, s_obs])
#         pca = PCA(n_components=latent_factor_size)
#         pca.fit(test)
#         X_train_pca = pca.transform(test)
#         init_values["cell_codebook"] = torch.tensor(pca.components_)
#         init_values["u_pcs_mean"] = torch.tensor(pca.mean_[: adata.shape[1]])
#         init_values["s_pcs_mean"] = torch.tensor(pca.mean_[adata.shape[1] :])
#         if plate_size == 2:
#             init_values["cell_code"] = (
#                 torch.tensor(X_train_pca).unsqueeze(-1).transpose(-1, -2)
#             )
#         else:
#             init_values["cell_code"] = torch.tensor(X_train_pca)

#     if num_aux_cells > 0:
#         np.random.seed(99)
#         if "cytotrace" in adata.obs.columns:
#             order_aux = np.array_split(
#                 np.sort(adata.obs["cytotrace"].values), 50
#             )
#             order_aux_list = []
#             for i in order_aux:
#                 order_aux_list.append(
#                     np.random.choice(
#                         np.where(
#                             (adata.obs["cytotrace"].values > i[0])
#                             & (adata.obs["cytotrace"].values < i[-1])
#                         )[0]
#                     )
#                 )
#             order_aux = np.array(order_aux_list)
#         else:
#             order_aux = np.argsort(
#                 (adata.layers["spliced"].toarray() > 0).sum(axis=1)
#             )[::-1]

#         init_values["order_aux"] = torch.from_numpy(
#             order_aux[:num_aux_cells].copy()
#         )

#         if input_type == "raw":
#             u_obs = adata.layers["raw_unspliced"].toarray()
#             s_obs = adata.layers["raw_spliced"].toarray()
#         init_values["aux_u_obs"] = torch.tensor(
#             u_obs[order_aux][:num_aux_cells]
#         )
#         init_values["aux_s_obs"] = torch.tensor(
#             s_obs[order_aux][:num_aux_cells]
#         )
#         init_values["cell_gene_state_aux"] = cell_gene_state[:num_aux_cells]
#         init_values["latent_time_aux"] = t[:num_aux_cells]

#         if latent_factor == "linear":
#             init_values["cell_code_aux"] = (
#                 torch.tensor(X_train_pca)
#                 .unsqueeze(-1)
#                 .transpose(-1, -2)[:num_aux_cells]
#             )

#     if shared_time:
#         cell_time = t.mean(dim=-1, keepdims=True)
#         init_values["cell_time"] = cell_time
#         init_values["latent_time"] = init_values["latent_time"] - cell_time

#         if num_aux_cells > 0:
#             init_values["cell_time_aux"] = cell_time[:num_aux_cells]
#             init_values["latent_time_aux"] = (
#                 init_values["latent_time_aux"] - init_values["cell_time_aux"]
#             )

#     for key in init_values:
#         print(key, init_values[key].shape)
#         print(init_values[key].isnan().sum())
#         assert init_values[key].isnan().sum() == 0
#     return init_values


# TODO: remove unused functions
# def trace(func):
#     def wrapper(*args, **kwargs):
#         params = list(args) + [f"{k}={v}" for k, v in kwargs.items()]
#         params_str = ",\n    ".join(map(str, params))
#         print(f"{func.__name__}(\n    {params_str},\n)")
#         return func(*args, **kwargs)

#     return wrapper


# def velocity_dus_dt(alpha, beta, gamma, tau, x):
#     """
#     Computes the velocity du/dt and ds/dt.

#     Args:
#         alpha (torch.Tensor): Alpha parameter.
#         beta (torch.Tensor): Beta parameter.
#         gamma (torch.Tensor): Gamma parameter.
#         tau (torch.Tensor): Time points.
#         x (Tuple[torch.Tensor, torch.Tensor]): Tuple containing u and s.

#     Returns:
#         Tuple[torch.Tensor, torch.Tensor]: Tuple containing du/dt and ds/dt.

#     Examples:
#         >>> import torch
#         >>> alpha = torch.tensor(0.5)
#         >>> beta = torch.tensor(0.4)
#         >>> gamma = torch.tensor(0.3)
#         >>> tau = torch.tensor(2.0)
#         >>> x = (torch.tensor(1.0), torch.tensor(0.5))
#         >>> velocity_dus_dt(alpha, beta, gamma, tau, x)
#         (tensor(0.1000), tensor(0.2500))
#     """
#     u, s = x
#     du_dt = alpha - beta * u
#     ds_dt = beta * u - gamma * s
#     return du_dt, ds_dt


# def rescale_time(dx_dt, t_start, t_end):
#     """
#     Converts an ODE to be solved on a batch of different time intervals into an
#     equivalent system of ODEs to be solved on [0, 1].

#     Args:
#         dx_dt (Callable): Function representing the ODE system.
#         t_start (torch.Tensor): Start time of the time interval.
#         t_end (torch.Tensor): End time of the time interval.

#     Returns:
#         Callable: Function representing the rescaled ODE system.

#     Examples:
#         >>> import torch
#         >>> def dx_dt(t, x):
#         ...     return -x
#         >>> t_start = torch.tensor(0.0)
#         >>> t_end = torch.tensor(1.0)
#         >>> rescaled_dx_dt = rescale_time(dx_dt, t_start, t_end)
#         >>> rescaled_dx_dt(torch.tensor(0.5), torch.tensor(2.0))
#         tensor(-2.)
#     """
#     dt = t_end - t_start

#     @functools.wraps(dx_dt)
#     def dx_ds(s, *args):
#         # move any batch dimensions in s to the left of all batch dimensions in dt
#         # s_shape = (1,) if len(s.shape) == 0 else s.shape  # XXX unnecessary?
#         s = s.reshape(s.shape + (1,) * len(dt.shape))
#         t = s * dt + t_start
#         xts = dx_dt(t, *args)
#         if isinstance(xts, tuple):
#             xss = tuple(xt * dt for xt in xts)
#         else:
#             xss = xts * dt
#         return xss

#     return dx_ds


# def ode_mrna_dynamics(tau, u0, s0, alpha, beta, gamma):
#     """
#     Solves the ODE system for mRNA dynamics.

#     Args:
#         tau (torch.Tensor): Time points.
#         u0 (torch.Tensor): Initial value of u.
#         s0 (torch.Tensor): Initial value of s.
#         alpha (torch.Tensor): Alpha parameter.
#         beta (torch.Tensor): Beta parameter.
#         gamma (torch.Tensor): Gamma parameter.

#     Returns:
#         Tuple[torch.Tensor, torch.Tensor]: Tuple containing the final values of u and s.
#     """
#     """
#     Examples:
#         >>> import torch
#         >>> tau = torch.tensor(2.0)
#         >>> u0 = torch.tensor(1.0)
#         >>> s0 = torch.tensor(0.5)
#         >>> alpha = torch.tensor(0.5)
#         >>> beta = torch.tensor(0.4)
#         >>> gamma = torch.tensor(0.3)
#         >>> ode_mrna_dynamics(tau, u0, s0, alpha, beta, gamma)
#         (tensor(0.6703), tensor(0.4596))
#     """
#     dx_dt = functools.partial(velocity_dus_dt, alpha, beta, gamma)
#     dx_dt = rescale_time(
#         dx_dt, torch.tensor(0.0, dtype=tau.dtype, device=tau.device), tau
#     )
#     grid = torch.linspace(0.0, 1.0, 100, dtype=tau.dtype, device=tau.device)
#     x0 = (u0.expand_as(tau), s0.expand_as(tau))
#     # xts = odeint_adjoint(dx_dt, x0, grid, adjoint_params=(alpha, beta, gamma, tau))
#     xts = odeint(dx_dt, x0, grid)
#     uts, sts = xts
#     ut, st = uts[-1], sts[-1]
#     return ut, st


# def get_velocity_samples(posterior_samples, model):
#     """
#     Computes the velocity samples from the posterior samples.

#     Args:
#         posterior_samples (dict): Dictionary containing posterior samples.
#         model: Model used for predictions.

#     Returns:
#         torch.Tensor: Velocity samples.

#     Examples:
#         >>> import torch
#         >>> posterior_samples = {
#         ...     "beta": torch.tensor([[0.4]]),
#         ...     "gamma": torch.tensor([[0.3]]),
#         ...     "u_scale": torch.tensor([[[1.0, 2.0]]]),
#         ...     "s_scale": torch.tensor([[[2.0, 4.0]]]),
#         ...     "u": torch.tensor([[1.0, 2.0]]),
#         ...     "s": torch.tensor([[0.5, 1.0]])
#         ... }
#         >>> model = None  # Model is not used in the function
#         >>> get_velocity_samples(posterior_samples, model)
#         tensor([[0.6500, 1.3000]])
#     """
#     beta = posterior_samples["beta"].mean(0)[0]
#     gamma = posterior_samples["gamma"].mean(0)[0]
#     scale = (
#         posterior_samples["u_scale"][:, 0, :]
#         / posterior_samples["s_scale"][:, 0, :]
#     ).mean(0)
#     ut = posterior_samples["u"] / scale
#     st = posterior_samples["s"]
#     v = beta * ut - gamma * st
#     return v

# def debug(x):
#     if torch.any(torch.isnan(x)):
#         print("nan number: ", torch.isnan(x).sum())
#         pdb.set_trace()


# def site_is_discrete(site: dict) -> bool:
#     return (
#         site["type"] == "sample"
#         and not site["is_observed"]
#         and getattr(site["fn"], "has_enumerate_support", False)
#     )


# def get_pylogger(name=__name__, log_level="DEBUG") -> logging.Logger:
#     """Initializes multi-GPU-friendly python command line logger."""

#     formatter = colorlog.ColoredFormatter(
#         "%(log_color)s%(levelname)s:%(name)s: %(message)s"
#         # "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s",
#         # datefmt=None,
#         # reset=True,
#         # log_colors={
#         #     "debug": "cyan",
#         #     "info": "green",
#         #     "warning": "yellow",
#         #     "error": "red",
#         #     "exception": "red",
#         #     "fatal": "red",
#         #     "critical": "red",
#         #     },
#     )

#     handler = colorlog.StreamHandler()
#     handler.setFormatter(formatter)
#     logger = colorlog.getLogger(name)
#     logger.setLevel(log_level)
#     logger.addHandler(handler)
#     logger.propagate = False

#     # this ensures all logging levels get marked with the rank zero decorator
#     # otherwise logs would get multiplied for each GPU process in multi-GPU setup
#     logging_levels = (
#         "debug",
#         "info",
#         "warning",
#         "error",
#         "exception",
#         "fatal",
#         "critical",
#     )
#     for level in logging_levels:
#         setattr(logger, level, rank_zero_only(getattr(logger, level)))

#     return logger
#     for level in logging_levels:
#         setattr(logger, level, rank_zero_only(getattr(logger, level)))

#     return logger



from typing import Tuple

import numpy as np
import pandas as pd
import scvelo as scv
import sklearn
import umap
from anndata import AnnData
from astropy.stats import rayleightest
from numpy import ndarray
from sklearn.pipeline import Pipeline

from pyrovelocity.cytotrace import compute_similarity2
from pyrovelocity.utils import ensure_numpy_array


def compute_mean_vector_field(
    posterior_samples,
    adata,
    basis="umap",
    n_jobs=1,
    spliced="spliced_pyro",
    raw=False,
):
    scv.pp.neighbors(adata, use_rep="pca")

    adata.var["velocity_genes"] = True

    if spliced == "spliced_pyro":
        if raw:
            ut = posterior_samples["ut"]
            st = posterior_samples["st"]
            ut = ut / ut.sum(axis=-1, keepdims=True)
            st = st / st.sum(axis=-1, keepdims=True)
        else:
            ut = posterior_samples["ut"]
            st = posterior_samples["st"]
        adata.layers["spliced_pyro"] = st.mean(0).squeeze()
        # if ('u_scale' in posterior_samples) and ('s_scale' in posterior_samples): # TODO: two scale for Normal distribution
        if (
            "u_scale" in posterior_samples
        ):  # only one scale for Poisson distribution
            adata.layers["velocity_pyro"] = (
                ut * posterior_samples["beta"] / posterior_samples["u_scale"]
                - st * posterior_samples["gamma"]
            ).mean(0)
        else:
            if "beta_k" in posterior_samples:
                adata.layers["velocity_pyro"] = (
                    (
                        ut * posterior_samples["beta_k"]
                        - posterior_samples["st"] * posterior_samples["gamma_k"]
                    )
                    .mean(0)
                    .squeeze()
                )
            else:
                adata.layers["velocity_pyro"] = (
                    ut * posterior_samples["beta"]
                    - posterior_samples["st"] * posterior_samples["gamma"]
                ).mean(0)
        scv.tl.velocity_graph(
            adata, vkey="velocity_pyro", xkey="spliced_pyro", n_jobs=n_jobs
        )
    elif spliced in ["Ms"]:
        ut = adata.layers["Mu"]
        st = adata.layers["Ms"]
        if ("u_scale" in posterior_samples) and (
            "s_scale" in posterior_samples
        ):
            adata.layers["velocity_pyro"] = (
                ut
                * posterior_samples["beta"]
                / (posterior_samples["u_scale"] / posterior_samples["s_scale"])
                - st * posterior_samples["gamma"]
            ).mean(0)
        else:
            adata.layers["velocity_pyro"] = (
                ut * posterior_samples["beta"]
                - posterior_samples["st"] * posterior_samples["gamma"]
            ).mean(0)
        scv.tl.velocity_graph(
            adata, vkey="velocity_pyro", xkey="Ms", n_jobs=n_jobs
        )
    elif spliced in ["spliced"]:
        ut = adata.layers["unspliced"]
        st = adata.layers["spliced"]
        if ("u_scale" in posterior_samples) and (
            "s_scale" in posterior_samples
        ):
            adata.layers["velocity_pyro"] = (
                ut
                * posterior_samples["beta"]
                / (posterior_samples["u_scale"] / posterior_samples["s_scale"])
                - st * posterior_samples["gamma"]
            ).mean(0)
        else:
            adata.layers["velocity_pyro"] = (
                ut * posterior_samples["beta"]
                - posterior_samples["st"] * posterior_samples["gamma"]
            ).mean(0)
        scv.tl.velocity_graph(
            adata, vkey="velocity_pyro", xkey="spliced", n_jobs=n_jobs
        )

    scv.tl.velocity_embedding(adata, vkey="velocity_pyro", basis=basis)


def compute_volcano_data(
    posterior_samples,
    adata,
    time_correlation_with="s",
    selected_genes=None,
    negative=False,
) -> None:
    assert isinstance(posterior_samples, (tuple, list))
    assert isinstance(adata, (tuple, list))
    assert "s" in posterior_samples[0]
    assert "alpha" in posterior_samples[0]

    maes_list = []
    cors = []
    genes = []
    labels = []
    switching = []
    for p, ad, label in zip(posterior_samples, adata, ["train", "valid"]):
        print(label)
        for sample in range(p["alpha"].shape[0]):
            maes_list.append(
                mae_per_gene(
                    p["s"][sample].squeeze(),
                    ensure_numpy_array(ad.layers["raw_spliced"]),
                )
            )
            df_genes_cors = compute_similarity2(
                p[time_correlation_with][sample].squeeze(),
                p["cell_time"][sample].squeeze().reshape(-1, 1),
            )
            cors.append(df_genes_cors[0])
            genes.append(ad.var_names.values)
            labels.append([f"Poisson_{label}"] * len(ad.var_names.values))

    volcano_data = pd.DataFrame(
        {
            "mean_mae": np.hstack(maes_list),
            "label": np.hstack(labels),
            "time_correlation": np.hstack(cors),
            "genes": np.hstack(genes),
        }
    )
    volcano_data = volcano_data.groupby("genes").mean(
        ["mean_mae", "time_correlation"]
    )

    volcano_data.loc[:, "mean_mae_rank"] = volcano_data.mean_mae.rank(
        ascending=False
    )
    volcano_data.loc[
        :, "time_correlation_rank"
    ] = volcano_data.time_correlation.apply(abs).rank(ascending=False)
    volcano_data.loc[:, "rank_product"] = (
        volcano_data.mean_mae_rank * volcano_data.time_correlation_rank
    )

    if selected_genes is None:
        genes = (
            volcano_data.sort_values("mean_mae", ascending=False)
            .head(300)
            .sort_values("time_correlation", ascending=negative)
            .head(4)
            .index
        )
    else:
        genes = selected_genes
    volcano_data.loc[:, "selected genes"] = 0
    volcano_data.loc[genes, "selected genes"] = 1

    return volcano_data, genes


def vector_field_uncertainty(
    adata: AnnData,
    posterior_samples: dict[str, ndarray],
    basis: str = "tsne",
    n_jobs: int = 1,
    denoised: bool = False,
) -> Tuple[ndarray, ndarray, ndarray]:
    """Run cosine similarity-based vector field across posterior samples"""

    # fig, ax = plt.subplots(10, 3)
    # fig.set_size_inches(16, 36)
    # ax = ax.flatten()
    v_map_all = []
    if ("u_scale" in posterior_samples) and (
        "s_scale" in posterior_samples
    ):  # Gaussian models
        scale = posterior_samples["u_scale"] / posterior_samples["s_scale"]
    elif ("u_scale" in posterior_samples) and not (
        "s_scale" in posterior_samples
    ):  # Poisson Model 2
        scale = posterior_samples["u_scale"]
    else:  # Poisson Model 1
        scale = 1

    if "beta_k" in posterior_samples:
        velocity_samples = (
            posterior_samples["ut"] * posterior_samples["beta_k"] / scale
            - posterior_samples["st"] * posterior_samples["gamma_k"]
        )
    else:
        velocity_samples = (
            posterior_samples["beta"] * posterior_samples["ut"] / scale
            - posterior_samples["gamma"] * posterior_samples["st"]
        )
    if denoised:
        projection = [
            (
                "PCA",
                sklearn.decomposition.PCA(random_state=99, n_components=50),
            ),
            ("UMAP", umap.UMAP(random_state=99, n_components=2)),
        ]
        pipelines = Pipeline(projection)
        expression = [posterior_samples["st"].mean(0)]
        pipelines.fit(expression[0])
        umap_orig = pipelines.transform(expression[0])
        adata.obsm["X_umap1"] = umap_orig
        joint_pcs = pipelines.steps[0][1].transform(expression[0])
        adata.obsm["X_pyropca"] = joint_pcs
        scv.pp.neighbors(adata, use_rep="pyropca")
    else:
        scv.pp.neighbors(adata, use_rep="pca")

    assert len(posterior_samples["st"].shape) == 3
    adata.var["velocity_genes"] = True
    for sample in range(posterior_samples["st"].shape[0]):
        adata.layers["spliced_pyro"] = posterior_samples["st"][sample]
        adata.layers["velocity_pyro"] = velocity_samples[sample]

        if basis == "pca":
            scv.pp.pca(adata)
            scv.tl.velocity_embedding(
                adata,
                vkey="velocity_pyro",
                basis="pca",
                direct_pca_projection=True,
            )
        else:
            scv.tl.velocity_graph(
                adata, vkey="velocity_pyro", xkey="spliced_pyro", n_jobs=n_jobs
            )
            scv.tl.velocity_embedding(adata, vkey="velocity_pyro", basis=basis)
        v_map_all.append(adata.obsm[f"velocity_pyro_{basis}"])
    v_map_all = np.stack(v_map_all)
    embeds_radian = np.arctan2(v_map_all[:, :, 1], v_map_all[:, :, 0])
    from statsmodels.stats.multitest import multipletests

    rayleightest_pval = rayleightest(embeds_radian, axis=-2)
    _, fdri, _, _ = multipletests(rayleightest_pval, method="fdr_bh")
    return v_map_all, embeds_radian, fdri


def mae_per_gene(pred_counts: ndarray, true_counts: ndarray) -> ndarray:
    """Computes mean average error between counts and predicted probabilities."""
    error = np.abs(true_counts - pred_counts).sum(-2)
    total = np.clip(true_counts.sum(-2), 1, np.inf)
    return -np.array(error / total)



from dataclasses import make_dataclass
from typing import Any, Dict, Tuple, Type

from mashumaro.mixins.json import DataClassJSONMixin

from pyrovelocity.data import download_dataset
from pyrovelocity.preprocess import preprocess_dataset
from pyrovelocity.train import train_dataset
from pyrovelocity.workflows.configuration import create_dataclass_from_callable

# These can be used to override the default values of the dataclass
# but are generally only required if the callable interface lacks
# complete and dataclass-compatible type annotations.
custom_types_defaults: Dict[str, Tuple[Type, Any]] = {
    # "data_set_name": (str, "pancreas"),
    # "source": (str, "pyrovelocity"),
    "data_external_path": (str, "data/external"),
}

download_dataset_fields = create_dataclass_from_callable(
    download_dataset, custom_types_defaults
)

DownloadDatasetInterface = make_dataclass(
    "DownloadDatasetInterface",
    download_dataset_fields,
    bases=(DataClassJSONMixin,),
)
DownloadDatasetInterface.__module__ = __name__


preprocess_data_types_defaults: Dict[str, Tuple[Type, Any]] = {
    "adata": (str, "data/external/simulated.h5ad"),
    "data_processed_path": (str, "data/processed"),
}

preprocess_data_fields = create_dataclass_from_callable(
    preprocess_dataset,
    preprocess_data_types_defaults,
)

PreprocessDataInterface = make_dataclass(
    "PreprocessDataInterface",
    preprocess_data_fields,
    bases=(DataClassJSONMixin,),
)
PreprocessDataInterface.__module__ = __name__

pyrovelocity_train_types_defaults: Dict[str, Tuple[Type, Any]] = {
    "adata": (str, "data/processed/simulated_processed.h5ad"),
    "use_gpu": (bool, False),
}

pyrovelocity_train_fields = create_dataclass_from_callable(
    train_dataset,
    pyrovelocity_train_types_defaults,
)

PyroVelocityTrainInterface = make_dataclass(
    "PyroVelocityTrainInterface",
    pyrovelocity_train_fields,
    bases=(DataClassJSONMixin,),
)
PyroVelocityTrainInterface.__module__ = __name__



from typing import Any, Dict, Optional, Tuple, Union

import torch
from pyro import poutine
from pyro.infer.autoguide import AutoLowRankMultivariateNormal, AutoNormal
from pyro.infer.autoguide.guides import AutoGuideList
from scvi.module.base import PyroBaseModuleClass

from pyrovelocity.models._velocity_model import VelocityModelAuto


class VelocityModule(PyroBaseModuleClass):
    """
    VelocityModule is an scvi-tools pyro module that combines the VelocityModelAuto and pyro AutoGuideList classes.

    Args:
        num_cells (int): Number of cells.
        num_genes (int): Number of genes.
        model_type (str, optional): Model type. Default is "auto".
        guide_type (str, optional): Guide type. Default is "velocity_auto".
        likelihood (str, optional): Likelihood type. Default is "Poisson".
        shared_time (bool, optional): If True, a shared time parameter will be used. Default is True.
        t_scale_on (bool, optional): If True, scale time parameter. Default is False.
        plate_size (int, optional): Size of the plate set. Default is 2.
        latent_factor (str, optional): Latent factor. Default is "none".
        latent_factor_operation (str, optional): Latent factor operation mode. Default is "selection".
        latent_factor_size (int, optional): Size of the latent factor. Default is 10.
        inducing_point_size (int, optional): Inducing point size. Default is 0.
        include_prior (bool, optional): If True, include prior in the model. Default is False.
        use_gpu (int, optional): GPU device index. Default is 0.
        num_aux_cells (int, optional): Number of auxiliary cells. Default is 0.
        only_cell_times (bool, optional): If True, only model cell times. Default is True.
        decoder_on (bool, optional): If True, use the decoder. Default is False.
        add_offset (bool, optional): If True, add offset to the model. Default is True.
        correct_library_size (Union[bool, str], optional): Library size correction method. Default is True.
        cell_specific_kinetics (Optional[str], optional): Cell-specific kinetics method. Default is None.
        kinetics_num (Optional[int], optional): Number of kinetics. Default is None.
        **initial_values: Initial values for the model parameters.

    Examples:
        >>> from scvi.module.base import PyroBaseModuleClass
        >>> from pyrovelocity.models._velocity_module import VelocityModule
        >>> num_cells = 10
        >>> num_genes = 20
        >>> velocity_module1 = VelocityModule(
        ...     num_cells, num_genes, model_type="auto",
        ...     guide_type="auto_t0_constraint", add_offset=False
        ... )
        -----------
        auto
        auto_t0_constraint
        >>> type(velocity_module1.model)
        <class 'pyrovelocity.models._velocity_model.VelocityModelAuto'>
        >>> type(velocity_module1.guide)
        <class 'pyro.infer.autoguide.guides.AutoGuideList'>
        >>> velocity_module2 = VelocityModule(
        ...     num_cells, num_genes, model_type="auto",
        ...     guide_type="auto", add_offset=True
        ... )
        -----------
        auto
        auto
        >>> type(velocity_module2.model)
        <class 'pyrovelocity.models._velocity_model.VelocityModelAuto'>
        >>> type(velocity_module2.guide)
        <class 'pyro.infer.autoguide.guides.AutoGuideList'>
    """

    def __init__(
        self,
        num_cells: int,
        num_genes: int,
        model_type: str = "auto",
        guide_type: str = "velocity_auto",
        likelihood: str = "Poisson",
        shared_time: bool = True,
        t_scale_on: bool = False,
        plate_size: int = 2,
        latent_factor: str = "none",
        latent_factor_operation: str = "selection",
        latent_factor_size: int = 10,
        inducing_point_size: int = 0,
        include_prior: bool = False,
        use_gpu: int = 0,
        num_aux_cells: int = 0,
        only_cell_times: bool = True,
        decoder_on: bool = False,
        add_offset: bool = True,
        correct_library_size: Union[bool, str] = True,
        cell_specific_kinetics: Optional[str] = None,
        kinetics_num: Optional[int] = None,
        **initial_values,
    ) -> None:
        super().__init__()
        self.num_cells = num_cells
        self.num_genes = num_genes
        self.model_type = model_type
        self.guide_type = guide_type
        self._model = None
        self.plate_size = plate_size
        self.num_aux_cells = num_aux_cells
        self.only_cell_times = only_cell_times
        print("-----------")
        print(self.model_type)
        print(self.guide_type)

        self.cell_specific_kinetics = cell_specific_kinetics

        self._model = VelocityModelAuto(
            self.num_cells,
            self.num_genes,
            likelihood,
            shared_time,
            t_scale_on,
            self.plate_size,
            latent_factor,
            latent_factor_operation=latent_factor_operation,
            latent_factor_size=latent_factor_size,
            include_prior=include_prior,
            num_aux_cells=num_aux_cells,
            only_cell_times=self.only_cell_times,
            decoder_on=decoder_on,
            add_offset=add_offset,
            correct_library_size=correct_library_size,
            guide_type=self.guide_type,
            cell_specific_kinetics=self.cell_specific_kinetics,
            **initial_values,
        )

        guide = AutoGuideList(
            self._model, create_plates=self._model.create_plates
        )
        guide.append(
            AutoNormal(
                poutine.block(
                    self._model,
                    expose=[
                        "cell_time",
                        "u_read_depth",
                        "s_read_depth",
                        "kinetics_prob",
                        "kinetics_weights",
                    ],
                ),
                init_scale=0.1,
            )
        )

        if add_offset:
            guide.append(
                AutoLowRankMultivariateNormal(
                    poutine.block(
                        self._model,
                        expose=[
                            "alpha",
                            "beta",
                            "gamma",
                            "dt_switching",
                            "t0",
                            "u_scale",
                            "s_scale",
                            "u_offset",
                            "s_offset",
                        ],
                    ),
                    rank=10,
                    init_scale=0.1,
                )
            )
        else:
            guide.append(
                AutoLowRankMultivariateNormal(
                    poutine.block(
                        self._model,
                        expose=[
                            "alpha",
                            "beta",
                            "gamma",
                            "dt_switching",
                            "t0",
                            "u_scale",
                            "s_scale",
                        ],
                    ),
                    rank=10,
                    init_scale=0.1,
                )
            )
        self._guide = guide

    @property
    def model(self) -> VelocityModelAuto:
        return self._model

    @property
    def guide(self) -> AutoGuideList:
        return self._guide

    @staticmethod
    def _get_fn_args_from_batch(
        tensor_dict: Dict[str, torch.Tensor]
    ) -> Tuple[
        Tuple[
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
        ],
        Dict[Any, Any],
    ]:
        u_obs = tensor_dict["U"]
        s_obs = tensor_dict["X"]
        u_log_library = tensor_dict["u_lib_size"]
        s_log_library = tensor_dict["s_lib_size"]
        u_log_library_mean = tensor_dict["u_lib_size_mean"]
        s_log_library_mean = tensor_dict["s_lib_size_mean"]
        u_log_library_scale = tensor_dict["u_lib_size_scale"]
        s_log_library_scale = tensor_dict["s_lib_size_scale"]
        ind_x = tensor_dict["ind_x"].long().squeeze()
        cell_state = tensor_dict.get("pyro_cell_state")
        time_info = tensor_dict.get("time_info")
        return (
            u_obs,
            s_obs,
            u_log_library,
            s_log_library,
            u_log_library_mean,
            s_log_library_mean,
            u_log_library_scale,
            s_log_library_scale,
            ind_x,
            cell_state,
            time_info,
        ), {}



import json
import os
import uuid
from pathlib import Path
from typing import Dict, Optional, Tuple

import matplotlib.pyplot as plt
import mlflow
import numpy as np
import torch
from anndata._core.anndata import AnnData
from beartype import beartype
from mlflow import MlflowClient
from numpy import ndarray

from pyrovelocity.data import load_anndata_from_path
from pyrovelocity.io.compressedpickle import CompressedPickle
from pyrovelocity.logging import configure_logging
from pyrovelocity.models import PyroVelocity
from pyrovelocity.utils import print_anndata, print_attributes

logger = configure_logging(__name__)


@beartype
def train_dataset(
    adata: str | AnnData,
    data_set_name: str = "simulated",
    model_identifier: str = "model2",
    guide_type: str = "auto",
    model_type: str = "auto",
    batch_size: int = -1,
    use_gpu: int | bool = False,
    likelihood: str = "Poisson",
    num_samples: int = 30,
    log_every: int = 100,
    patient_improve: float = 1e-4,
    patient_init: int = 45,
    seed: int = 99,
    learning_rate: float = 0.01,
    max_epochs: int = 3000,
    include_prior: bool = True,
    library_size: bool = True,
    offset: bool = True,
    input_type: str = "raw",
    cell_specific_kinetics: Optional[str] = None,
    kinetics_num: int = 2,
    force: bool = False,
) -> Tuple[Path, Path, Path, Path, Path, Path]:
    """
    Loads processed data, trains model, and saves model and posterior samples.

    Inputs:
        "data/processed/{data_set_name}_processed.h5ad"

    Outputs:
        data:
            "models/{data_model}/trained.h5ad"
            "models/{data_model}/pyrovelocity.pkl"
            "models/{data_model}/posterior_samples.pkl.zst"
        models:
            models/{data_model}/model/
            ├── attr.pkl
            ├── model_params.pt
            ├── param_store_test.pt
            └── var_names.csv

    Args:
        data_set_name (str, optional): Data set name. Defaults to "simulated".
        model_identifier (str, optional): Model identifier. Defaults to "model1".
        force (bool, optional): Overwrite existing output. Defaults to False.

    Returns:
        Tuple[Path, Path, Path, Path, Path, Path]: Paths to saved data.

    Examples:
        >>> train_dataset() # xdoctest: +SKIP
    """
    ###########
    # load data
    ###########

    data_model = f"{data_set_name}_{model_identifier}"
    model_dir = Path(f"models/{data_model}")

    trained_data_path = model_dir / "trained.h5ad"
    model_path = model_dir / "model"
    posterior_samples_path = model_dir / "posterior_samples.pkl.zst"
    metrics_path = model_dir / "metrics.json"
    run_info_path = model_dir / "run_info.json"
    loss_plot_path = model_dir / "ELBO.png"

    logger.info(f"\n\nTraining: {data_model}\n\n")

    logger.info(
        f"\n\nVerifying existence of paths for:\n\n"
        f"  model data: {model_dir}\n"
    )
    Path(model_dir).mkdir(parents=True, exist_ok=True)

    #############
    # train model
    #############

    if torch.cuda.is_available():
        accelerators = list(range(torch.cuda.device_count()))
        use_gpu = accelerators.pop()
    else:
        use_gpu = False

    logger.info(f"GPU ID: {use_gpu}")

    if (
        os.path.isfile(trained_data_path)
        and os.path.exists(model_path)
        and os.path.isfile(posterior_samples_path)
        and not force
    ):
        logger.info(
            f"\n{trained_data_path}\n"
            f"{model_path}\n"
            f"{posterior_samples_path}\n"
            "all exist, set `force=True` to overwrite."
        )
        return (
            trained_data_path,
            model_path,
            posterior_samples_path,
            metrics_path,
            run_info_path,
            loss_plot_path,
        )
    else:
        logger.info(f"Training model: {data_model}")

        # UPDATE: v2.1.1 12/26/2022 autolog only supports pytorch lightning
        # mlflow.pytorch.autolog(log_every_n_epoch=200, log_models=False, silent=False)
        with mlflow.start_run(
            run_name=f"{data_model}-{uuid.uuid4().hex[:7]}"
        ) as run:
            mlflow.set_tag(
                "mlflow.runName", f"{data_model}-{run.info.run_id[:7]}"
            )
            print(f"Active run_id: {run.info.run_id}")
            mlflow.log_params(
                {
                    "data_set_name": data_set_name,
                    "model_identifier": model_identifier,
                    "guide_type": guide_type,
                    "model_type": model_type,
                    "batch_size": batch_size,
                }
            )

            adata, trained_model, posterior_samples = train_model(
                adata=adata,
                guide_type=guide_type,
                model_type=model_type,
                batch_size=batch_size,
                use_gpu=use_gpu,
                likelihood=likelihood,
                num_samples=num_samples,
                log_every=log_every,
                patient_improve=patient_improve,
                patient_init=patient_init,
                seed=seed,
                learning_rate=learning_rate,
                max_epochs=max_epochs,
                include_prior=include_prior,
                library_size=library_size,
                offset=offset,
                input_type=input_type,
                cell_specific_kinetics=cell_specific_kinetics,
                kinetics_num=kinetics_num,
                loss_plot_path=str(loss_plot_path),
            )

            logger.info("Data attributes after model training")

            run_id = run.info.run_id

        logger.info(f"Saving posterior samples: {posterior_samples_path}\n")
        CompressedPickle.save(
            posterior_samples_path,
            posterior_samples,
        )

        ##############
        # save metrics
        ##############

        r = mlflow.get_run(run_id)

        Path(metrics_path).write_text(json.dumps(r.data.metrics, indent=4))

        Path(run_info_path).write_text(
            json.dumps(r.to_dictionary()["info"], indent=4)
        )

        log_run_info(r)

        #############
        # postprocess
        #############

        if "pancreas" in data_model:
            logger.info("checking shared time")

            def check_shared_time(posterior_samples, adata):
                adata.obs["cell_time"] = (
                    posterior_samples["cell_time"].squeeze().mean(0)
                )
                adata.obs["1-Cytotrace"] = 1 - adata.obs["cytotrace"]

            check_shared_time(posterior_samples, adata)

        print_attributes(adata)
        print_anndata(adata)

        ##################
        # save checkpoints
        ##################

        logger.info(f"Saving trained data: {trained_data_path}")
        adata.write(trained_data_path)

        logger.info(f"Saving model: {model_path}")
        trained_model.save_model(model_path, overwrite=True)

        logger.info(
            f"\nReturning paths to saved data:\n\n"
            f"{trained_data_path}\n"
            f"{model_path}\n"
            f"{posterior_samples_path}\n"
            f"{metrics_path}\n"
            f"{run_info_path}\n"
            f"{loss_plot_path}\n"
        )
        return (
            trained_data_path,
            model_path,
            posterior_samples_path,
            metrics_path,
            run_info_path,
            loss_plot_path,
        )


@beartype
def train_model(
    adata: str | AnnData,
    guide_type: str = "auto",
    model_type: str = "auto",
    batch_size: int = -1,
    use_gpu: int | bool = False,
    likelihood: str = "Poisson",
    num_samples: int = 30,
    log_every: int = 100,
    patient_improve: float = 1e-4,
    patient_init: int = 45,
    seed: int = 99,
    learning_rate: float = 0.01,
    max_epochs: int = 3000,
    include_prior: bool = True,
    library_size: bool = True,
    offset: bool = True,
    input_type: str = "raw",
    cell_specific_kinetics: Optional[str] = None,
    kinetics_num: int = 2,
    loss_plot_path: str = "loss_plot.png",
) -> Tuple[AnnData, PyroVelocity, Dict[str, ndarray]]:
    """
    Train a PyroVelocity model to provide probabilistic estimates of RNA velocity
    for single-cell RNA sequencing data with quantified splice variants.

    Args:
        adata (str | AnnData): Path to a file that can be read to an AnnData object or an AnnData object.
        guide_type (str, optional): The type of guide function for the Pyro model. Default is "auto".
        model_type (str, optional): The type of Pyro model. Default is "auto".
        batch_size (int, optional): Batch size for training. Default is -1, which indicates using the full dataset.
        use_gpu (int, optional): Whether to use GPU for training. Default is 0, which indicates not using GPU.
        likelihood (str, optional): Likelihood function for the Pyro model. Default is "Poisson".
        num_samples (int, optional): Number of posterior samples. Default is 30.
        log_every (int, optional): Frequency of logging progress. Default is 100.
        patient_improve (float, optional): Minimum improvement in training loss for early stopping. Default is 5e-4.
        patient_init (int, optional): Number of initial training epochs before early stopping is enabled. Default is 30.
        seed (int, optional): Random seed for reproducibility. Default is 99.
        learning_rate (float, optional): Learning rate for the optimizer. Default is 0.01.
        max_epochs (int, optional): Maximum number of training epochs. Default is 3000.
        include_prior (bool, optional): Whether to include prior information in the model. Default is True.
        library_size (bool, optional): Whether to correct for library size. Default is True.
        offset (bool, optional): Whether to add an offset to the model. Default is False.
        input_type (str, optional): Type of input data. Default is "raw".
        cell_specific_kinetics (Optional[str], optional): Name of the attribute containing cell-specific kinetics information. Default is None.
        kinetics_num (int, optional): Number of kinetics parameters. Default is 2.
        loss_plot_path (str, optional): Path to save the loss plot. Default is "loss_plot.png".

        These arguments are deprecated:
        # svi_train: bool = False,
        # train_size: float = 1.0,
        # cell_state: str = "clusters",
        # svi_train (bool, optional): Whether to use Stochastic Variational Inference for training. Default is False.
        # train_size (float, optional): Proportion of data to be used for training. Default is 1.0.
        # cell_state (str, optional): Cell state attribute in the AnnData object. Default is "clusters".

    Returns:
        Tuple[PyroVelocity, Dict[str, ndarray]]: A tuple containing the trained PyroVelocity model and a dictionary of posterior samples.

    Examples:
        >>> from pyrovelocity.train import train_model
        >>> from pyrovelocity.utils import generate_sample_data
        >>> from pyrovelocity.preprocess import copy_raw_counts
        >>> adata = generate_sample_data(random_seed=99)
        >>> copy_raw_counts(adata)
        >>> _, model, posterior_samples = train_model(adata, use_gpu=False, seed=99, max_epochs=200, loss_plot_path="loss_plot_docs.png")
    """
    if isinstance(adata, str):
        adata = load_anndata_from_path(adata)

    print_anndata(adata)
    print_attributes(adata)

    PyroVelocity.setup_anndata(adata)

    model = PyroVelocity(
        adata,
        likelihood=likelihood,
        model_type=model_type,
        guide_type=guide_type,
        correct_library_size=library_size,
        add_offset=offset,
        include_prior=include_prior,
        input_type=input_type,
        cell_specific_kinetics=cell_specific_kinetics,
        kinetics_num=kinetics_num,
    )

    if batch_size == -1:
        batch_size = adata.shape[0]

    if batch_size >= adata.shape[0]:
        losses = model.train_faster(
            max_epochs=max_epochs,
            lr=learning_rate,
            use_gpu=use_gpu,
            seed=seed,
            patient_improve=patient_improve,
            patient_init=patient_init,
            log_every=log_every,
        )
    else:
        losses = model.train_faster_with_batch(
            max_epochs=max_epochs,
            batch_size=batch_size,
            log_every=log_every,
            lr=learning_rate,
            use_gpu=use_gpu,
            seed=seed,
            patient_improve=patient_improve,
            patient_init=patient_init,
        )
    fig, ax = plt.subplots()
    fig.set_size_inches(2.5, 1.5)
    ax.scatter(
        np.arange(len(losses)),
        -np.array(losses),
        label="train",
        alpha=0.25,
    )
    set_loss_plot_axes(ax)
    posterior_samples = model.generate_posterior_samples(
        model.adata, num_samples=num_samples, batch_size=512
    )

    fig.savefig(loss_plot_path, facecolor="white", bbox_inches="tight")

    return adata, model, posterior_samples


def set_loss_plot_axes(ax):
    ax.set_yscale("symlog")
    ax.set_xlabel("Epochs")
    ax.set_ylabel("-ELBO")


def log_run_info(r: mlflow.entities.run.Run) -> None:
    tags = {k: v for k, v in r.data.tags.items() if not k.startswith("mlflow.")}
    artifacts = [
        f.path for f in MlflowClient().list_artifacts(r.info.run_id, "model")
    ]
    logger.info(
        f"\nrun_id: {r.info.run_id}\n"
        f"artifacts: {artifacts}\n"
        f"params: {r.data.params}\n"
        f"metrics: {r.data.metrics}\n"
        f"tags: {tags}\n"
    )



from dataclasses import asdict
from datetime import timedelta

from flytekit import Resources, dynamic, task, workflow
from flytekit.extras.accelerators import T4
from flytekit.types.directory import FlyteDirectory
from flytekit.types.file import FlyteFile

from pyrovelocity.data import download_dataset
from pyrovelocity.interfaces import (
    DownloadDatasetInterface,
    PreprocessDataInterface,
    PyroVelocityTrainInterface,
)
from pyrovelocity.logging import configure_logging
from pyrovelocity.preprocess import preprocess_dataset
from pyrovelocity.train import train_dataset
from pyrovelocity.workflows.main_configuration import (
    ResourcesJSON,
    TrainingOutputs,
    default_training_resources,
    larry_configuration,
    pancreas_configuration,
    pbmc68k_configuration,
    pons_configuration,
    simulated_configuration,
)

# from pyrovelocity.workflows.main_configuration import (
#     test_pancreas_configuration as pancreas_configuration,
# )
# from pyrovelocity.workflows.main_configuration import (
#     test_simulated_configuration as simulated_configuration,
# )

logger = configure_logging(__name__)

cache_version = "0.2.0b9"


@task(
    cache=True,
    cache_version=cache_version,
    retries=3,
    interruptible=True,
    timeout=timedelta(minutes=20),
    requests=Resources(cpu="2", mem="4Gi", ephemeral_storage="16Gi"),
    limits=Resources(cpu="8", mem="16Gi", ephemeral_storage="200Gi"),
)
def download_data(download_dataset_args: DownloadDatasetInterface) -> FlyteFile:
    """
    Download external data.
    """
    dataset_path = download_dataset(**asdict(download_dataset_args))
    return FlyteFile(path=dataset_path)


@task(
    cache=True,
    cache_version=cache_version,
    retries=3,
    interruptible=False,
    timeout=timedelta(minutes=60),
    requests=Resources(cpu="8", mem="30Gi", ephemeral_storage="32Gi"),
    limits=Resources(cpu="16", mem="60Gi", ephemeral_storage="200Gi"),
)
def preprocess_data(
    data: FlyteFile, preprocess_data_args: PreprocessDataInterface
) -> FlyteFile:
    """
    Download external data.
    """
    data_path = data.download()
    print(f"Flyte preprocess input data path: {data_path}")
    preprocess_data_args.adata = str(data_path)
    _, processed_dataset_path = preprocess_dataset(
        **asdict(preprocess_data_args),
    )
    return FlyteFile(path=processed_dataset_path)


@task(
    cache=True,
    cache_version=cache_version,
    retries=3,
    interruptible=False,
    timeout=timedelta(minutes=120),
    container_image="{{.image.gpu.fqn}}:{{.image.gpu.version}}",
    requests=Resources(cpu="8", mem="30Gi", ephemeral_storage="50Gi", gpu="1"),
    limits=Resources(cpu="16", mem="60Gi", ephemeral_storage="200Gi", gpu="1"),
    accelerator=T4,
)
def train_model(
    data: FlyteFile,
    train_model_configuration: PyroVelocityTrainInterface,
) -> TrainingOutputs:
    """
    Train model.
    """
    data_path = data.download()
    print(f"Flyte train model input data path: {data_path}")
    train_model_configuration.adata = str(data_path)
    (
        trained_data_path,
        model_path,
        posterior_samples_path,
        metrics_path,
        run_info_path,
        loss_plot_path,
    ) = train_dataset(
        **asdict(train_model_configuration),
    )
    return TrainingOutputs(
        trained_data_path=FlyteFile(path=trained_data_path),
        model_path=FlyteDirectory(path=model_path),
        posterior_samples_path=FlyteFile(path=posterior_samples_path),
        metrics_path=FlyteFile(path=metrics_path),
        run_info_path=FlyteFile(path=run_info_path),
        loss_plot_path=FlyteFile(path=loss_plot_path),
    )


@dynamic
def module_workflow(
    download_dataset_args: DownloadDatasetInterface = DownloadDatasetInterface(),
    preprocess_data_args: PreprocessDataInterface = PreprocessDataInterface(),
    train_model_configurations: list[PyroVelocityTrainInterface] = [
        PyroVelocityTrainInterface()
    ],
    train_model_resources: list[ResourcesJSON] = default_training_resources,
) -> list[TrainingOutputs]:
    """
    The module workflow is applied to a single dataset together with a list of
    model configurations.

    There are three ways to execute the train_model() task, which impacts how
    subsequent tasks are executed. This can also be executed with the @workflow
    decorator for a single model configuration

    ```python
    train_model_configuration: PyroVelocityTrainInterface = PyroVelocityTrainInterface(),
    ) -> TrainingOutputs:
    ...
    model_outputs = train_model(
        data=processed_data,
        train_model_configuration=train_model_configuration,
    )
    ```

    or for multiple model configurations using a map task

    ```python
    import functools
    from flytekit.experimental import map_task

    partial_train_model = functools.partial(train_model, data=processed_data)
    model_outputs = map_task(partial_train_model)(
        train_model_args=train_model_configurations,
    )
    ```

    The dynamic workflow is preferred to support both mapping of tasks over
    multiple model configurations and overriding resources from configuration
    data.
    """
    data = download_data(download_dataset_args=download_dataset_args)
    processed_data = preprocess_data(
        data=data, preprocess_data_args=preprocess_data_args
    )

    model_outputs: list[TrainingOutputs] = list()
    for train_model_configuration in train_model_configurations:
        model_output = train_model(
            data=processed_data,
            train_model_configuration=train_model_configuration,
        ).with_overrides(
            requests=Resources(**asdict(train_model_resources[0])),
            limits=Resources(**asdict(train_model_resources[1])),
        )
        model_outputs.append(model_output)
    return model_outputs


@workflow
def training_workflow(
    simulated_dataset_configuration: DownloadDatasetInterface = simulated_configuration.download_dataset,
    simulated_preprocess_configuration: PreprocessDataInterface = simulated_configuration.preprocess_data,
    simulated_train_model_configurations: list[
        PyroVelocityTrainInterface
    ] = simulated_configuration.training_configurations,
    simulated_train_model_resources: list[
        ResourcesJSON
    ] = simulated_configuration.training_resources,
    pancreas_dataset_configuration: DownloadDatasetInterface = pancreas_configuration.download_dataset,
    pancreas_preprocess_configuration: PreprocessDataInterface = pancreas_configuration.preprocess_data,
    pancreas_train_model_configurations: list[
        PyroVelocityTrainInterface
    ] = pancreas_configuration.training_configurations,
    pancreas_train_model_resources: list[
        ResourcesJSON
    ] = pancreas_configuration.training_resources,
    pbmc68k_dataset_configuration: DownloadDatasetInterface = pbmc68k_configuration.download_dataset,
    pbmc68k_preprocess_configuration: PreprocessDataInterface = pbmc68k_configuration.preprocess_data,
    pbmc68k_train_model_configurations: list[
        PyroVelocityTrainInterface
    ] = pbmc68k_configuration.training_configurations,
    pbmc68k_train_model_resources: list[
        ResourcesJSON
    ] = pbmc68k_configuration.training_resources,
    pons_dataset_configuration: DownloadDatasetInterface = pons_configuration.download_dataset,
    pons_preprocess_configuration: PreprocessDataInterface = pons_configuration.preprocess_data,
    pons_train_model_configurations: list[
        PyroVelocityTrainInterface
    ] = pons_configuration.training_configurations,
    pons_train_model_resources: list[
        ResourcesJSON
    ] = pons_configuration.training_resources,
    larry_dataset_configuration: DownloadDatasetInterface = larry_configuration.download_dataset,
    larry_preprocess_configuration: PreprocessDataInterface = larry_configuration.preprocess_data,
    larry_train_model_configurations: list[
        PyroVelocityTrainInterface
    ] = larry_configuration.training_configurations,
    larry_train_model_resources: list[
        ResourcesJSON
    ] = larry_configuration.training_resources,
) -> list[list[TrainingOutputs]]:
    """
    Apply the module_workflow to a collection of datasets.

    TODO: Update interface extraction to support nested dataclasses, which will
    allow simplification of input arguments to:

    simulated_configuration: WorkflowConfiguration = simulated_configuration,
    """
    simulated = module_workflow(
        download_dataset_args=simulated_dataset_configuration,
        preprocess_data_args=simulated_preprocess_configuration,
        train_model_configurations=simulated_train_model_configurations,
        train_model_resources=simulated_train_model_resources,
    )

    pancreas = module_workflow(
        download_dataset_args=pancreas_dataset_configuration,
        preprocess_data_args=pancreas_preprocess_configuration,
        train_model_configurations=pancreas_train_model_configurations,
        train_model_resources=pancreas_train_model_resources,
    )

    pbmc68k = module_workflow(
        download_dataset_args=pbmc68k_dataset_configuration,
        preprocess_data_args=pbmc68k_preprocess_configuration,
        train_model_configurations=pbmc68k_train_model_configurations,
        train_model_resources=pbmc68k_train_model_resources,
    )

    pons = module_workflow(
        download_dataset_args=pons_dataset_configuration,
        preprocess_data_args=pons_preprocess_configuration,
        train_model_configurations=pons_train_model_configurations,
        train_model_resources=pons_train_model_resources,
    )

    larry = module_workflow(
        download_dataset_args=larry_dataset_configuration,
        preprocess_data_args=larry_preprocess_configuration,
        train_model_configurations=larry_train_model_configurations,
        train_model_resources=larry_train_model_resources,
    )

    return [
        simulated,
        pancreas,
        pbmc68k,
        pons,
        larry,
    ]


if __name__ == "__main__":
    print(f"Running module_workflow() { module_workflow() }")
    print(f"Running training_workflow() { training_workflow() }")



from typing import Tuple

import torch
from beartype import beartype


@beartype
def mrna_dynamics(
    tau: torch.Tensor,
    u0: torch.Tensor,
    s0: torch.Tensor,
    alpha: torch.Tensor,
    beta: torch.Tensor,
    gamma: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Computes the mRNA dynamics given the parameters and initial conditions.

    `st_gamma_equals_beta` is taken from On the Mathematics of RNA Velocity I:
    Theoretical Analysis: Equation (2.12) when gamma == beta

    Args:
        tau (torch.Tensor): Time points.
        u0 (torch.Tensor): Initial value of u.
        s0 (torch.Tensor): Initial value of s.
        alpha (torch.Tensor): Alpha parameter.
        beta (torch.Tensor): Beta parameter.
        gamma (torch.Tensor): Gamma parameter.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Tuple containing the final values of u and s.

    Examples:
        >>> import torch
        >>> tau = torch.tensor(2.0)
        >>> u0 = torch.tensor(1.0)
        >>> s0 = torch.tensor(0.5)
        >>> alpha = torch.tensor(0.5)
        >>> beta = torch.tensor(0.4)
        >>> gamma = torch.tensor(0.3)
        >>> mrna_dynamics(tau, u0, s0, alpha, beta, gamma)
        (tensor(1.1377), tensor(0.9269))
    """
    expu, exps = torch.exp(-beta * tau), torch.exp(-gamma * tau)

    expus = (alpha - u0 * beta) * inv(gamma - beta) * (exps - expu)

    ut = u0 * expu + alpha / beta * (1 - expu)
    st = s0 * exps + alpha / gamma * (1 - exps) + expus
    st_gamma_equals_beta = (
        s0 * expu + alpha / beta * (1 - expu) - (alpha - beta * u0) * tau * expu
    )
    st = torch.where(torch.isclose(gamma, beta), st_gamma_equals_beta, st)

    return ut, st


@beartype
def inv(x: torch.Tensor) -> torch.Tensor:
    """
    Computes the element-wise reciprocal of a tensor.

    Args:
        x (torch.Tensor): Input tensor.

    Returns:
        torch.Tensor: Tensor with element-wise reciprocal of x.

    Examples:
        >>> import torch
        >>> x = torch.tensor([2., 4., 0.5])
        >>> inv(x)
        tensor([0.5000, 0.2500, 2.0000])
    """
    return x.reciprocal()



from dataclasses import dataclass

from flytekit.types.directory import FlyteDirectory
from flytekit.types.file import FlyteFile
from mashumaro.mixins.json import DataClassJSONMixin

from pyrovelocity.interfaces import (
    DownloadDatasetInterface,
    PreprocessDataInterface,
    PyroVelocityTrainInterface,
)
from pyrovelocity.logging import configure_logging

logger = configure_logging(__name__)

UNIFORM_MAX_EPOCHS = 2000


@dataclass
class ResourcesJSON(DataClassJSONMixin):
    cpu: str
    mem: str
    gpu: str
    ephemeral_storage: str


testing_training_resources = [
    ResourcesJSON(
        cpu="4",
        mem="8Gi",
        gpu="1",
        ephemeral_storage="32Gi",
    ),
    ResourcesJSON(
        cpu="4",
        mem="8Gi",
        gpu="1",
        ephemeral_storage="32Gi",
    ),
]

default_training_resources = [
    ResourcesJSON(
        cpu="8",
        mem="30Gi",
        gpu="1",
        ephemeral_storage="50Gi",
    ),
    ResourcesJSON(
        cpu="16",
        mem="60Gi",
        gpu="1",
        ephemeral_storage="200Gi",
    ),
]

large_training_resources = [
    ResourcesJSON(
        cpu="16",
        mem="120Gi",
        gpu="1",
        ephemeral_storage="50Gi",
    ),
    ResourcesJSON(
        cpu="32",
        mem="200Gi",
        gpu="1",
        ephemeral_storage="200Gi",
    ),
]


@dataclass
class WorkflowConfiguration(DataClassJSONMixin):
    download_dataset: DownloadDatasetInterface
    preprocess_data: PreprocessDataInterface
    training_configurations: list[PyroVelocityTrainInterface]
    training_resources: list[ResourcesJSON]


@dataclass
class TrainingOutputs(DataClassJSONMixin):
    trained_data_path: FlyteFile
    model_path: FlyteDirectory
    posterior_samples_path: FlyteFile
    metrics_path: FlyteFile
    run_info_path: FlyteFile
    loss_plot_path: FlyteFile


simulated_dataset_args = DownloadDatasetInterface(
    data_set_name="simulated",
    source="simulate",
    n_obs=3000,
    n_vars=2000,
)
simulated_preprocess_data_args = PreprocessDataInterface(
    data_set_name=f"{simulated_dataset_args.data_set_name}",
    adata=f"{simulated_dataset_args.data_external_path}/{simulated_dataset_args.data_set_name}.h5ad",
)
simulated_train_model1_args = PyroVelocityTrainInterface(
    adata=f"{simulated_preprocess_data_args.data_processed_path}/{simulated_dataset_args.data_set_name}_processed.h5ad",
    data_set_name=f"{simulated_dataset_args.data_set_name}",
    model_identifier="model1",
    guide_type="auto_t0_constraint",
    offset=False,
    max_epochs=UNIFORM_MAX_EPOCHS,
)
simulated_train_model2_args = PyroVelocityTrainInterface(
    adata=f"{simulated_preprocess_data_args.data_processed_path}/{simulated_dataset_args.data_set_name}_processed.h5ad",
    data_set_name=f"{simulated_dataset_args.data_set_name}",
    model_identifier="model2",
    max_epochs=UNIFORM_MAX_EPOCHS,
)

simulated_configuration = WorkflowConfiguration(
    download_dataset=simulated_dataset_args,
    preprocess_data=simulated_preprocess_data_args,
    training_configurations=[
        simulated_train_model1_args,
        simulated_train_model2_args,
    ],
    # training_resources=testing_training_resources,
    training_resources=default_training_resources,
)

test_simulated_configuration = simulated_configuration
test_simulated_configuration.download_dataset.n_obs = 200
test_simulated_configuration.download_dataset.n_vars = 100
test_simulated_configuration.training_configurations[0].max_epochs = 300
test_simulated_configuration.training_configurations[1].max_epochs = 300

pancreas_dataset_args = DownloadDatasetInterface(
    data_set_name="pancreas",
)
pancreas_preprocess_data_args = PreprocessDataInterface(
    data_set_name=f"{pancreas_dataset_args.data_set_name}",
    adata=f"{pancreas_dataset_args.data_external_path}/{pancreas_dataset_args.data_set_name}.h5ad",
    process_cytotrace=True,
)
pancreas_train_model1_args = PyroVelocityTrainInterface(
    adata=f"{pancreas_preprocess_data_args.data_processed_path}/{pancreas_dataset_args.data_set_name}_processed.h5ad",
    data_set_name=f"{pancreas_dataset_args.data_set_name}",
    model_identifier="model1",
    guide_type="auto_t0_constraint",
    offset=False,
    max_epochs=UNIFORM_MAX_EPOCHS,
)
pancreas_train_model2_args = PyroVelocityTrainInterface(
    adata=f"{pancreas_preprocess_data_args.data_processed_path}/{pancreas_dataset_args.data_set_name}_processed.h5ad",
    data_set_name=f"{pancreas_dataset_args.data_set_name}",
    model_identifier="model2",
    max_epochs=UNIFORM_MAX_EPOCHS,
)

pancreas_configuration = WorkflowConfiguration(
    download_dataset=pancreas_dataset_args,
    preprocess_data=pancreas_preprocess_data_args,
    training_configurations=[
        pancreas_train_model1_args,
        pancreas_train_model2_args,
    ],
    # training_resources=testing_training_resources,
    training_resources=default_training_resources,
)

test_pancreas_configuration = pancreas_configuration
test_pancreas_configuration.download_dataset.n_obs = 200
test_pancreas_configuration.download_dataset.n_vars = 500
test_pancreas_configuration.training_configurations[0].max_epochs = 300
test_pancreas_configuration.training_configurations[1].max_epochs = 300

pbmc68k_dataset_args = DownloadDatasetInterface(
    data_set_name="pbmc68k",
    # n_obs=500,
)
pbmc68k_preprocess_data_args = PreprocessDataInterface(
    data_set_name=f"{pbmc68k_dataset_args.data_set_name}",
    adata=f"{pbmc68k_dataset_args.data_external_path}/{pbmc68k_dataset_args.data_set_name}.h5ad",
    default_velocity_mode="stochastic",
    vector_field_basis="tsne",
)
pbmc68k_train_model1_args = PyroVelocityTrainInterface(
    adata=f"{pbmc68k_preprocess_data_args.data_processed_path}/{pbmc68k_dataset_args.data_set_name}_processed.h5ad",
    data_set_name=f"{pbmc68k_dataset_args.data_set_name}",
    model_identifier="model1",
    guide_type="auto_t0_constraint",
    offset=False,
    # cell_state="celltype",
    max_epochs=UNIFORM_MAX_EPOCHS,
)
pbmc68k_train_model2_args = PyroVelocityTrainInterface(
    adata=f"{pbmc68k_preprocess_data_args.data_processed_path}/{pbmc68k_dataset_args.data_set_name}_processed.h5ad",
    data_set_name=f"{pbmc68k_dataset_args.data_set_name}",
    model_identifier="model2",
    # cell_state="celltype",
    max_epochs=UNIFORM_MAX_EPOCHS,
)

pbmc68k_configuration = WorkflowConfiguration(
    download_dataset=pbmc68k_dataset_args,
    preprocess_data=pbmc68k_preprocess_data_args,
    training_configurations=[
        pbmc68k_train_model1_args,
        pbmc68k_train_model2_args,
    ],
    # training_resources=testing_training_resources,
    training_resources=default_training_resources,
)

pons_dataset_args = DownloadDatasetInterface(
    data_set_name="pons",
    # n_obs=200,
    # n_vars=500,
)
pons_preprocess_data_args = PreprocessDataInterface(
    data_set_name=f"{pons_dataset_args.data_set_name}",
    adata=f"{pons_dataset_args.data_external_path}/{pons_dataset_args.data_set_name}.h5ad",
)
pons_train_model1_args = PyroVelocityTrainInterface(
    adata=f"{pons_preprocess_data_args.data_processed_path}/{pons_dataset_args.data_set_name}_processed.h5ad",
    data_set_name=f"{pons_dataset_args.data_set_name}",
    model_identifier="model1",
    guide_type="auto_t0_constraint",
    offset=False,
    # cell_state="celltype",
    max_epochs=UNIFORM_MAX_EPOCHS,
)
pons_train_model2_args = PyroVelocityTrainInterface(
    adata=f"{pons_preprocess_data_args.data_processed_path}/{pons_dataset_args.data_set_name}_processed.h5ad",
    data_set_name=f"{pons_dataset_args.data_set_name}",
    model_identifier="model2",
    # cell_state="celltype",
    max_epochs=UNIFORM_MAX_EPOCHS,
)

pons_configuration = WorkflowConfiguration(
    download_dataset=pons_dataset_args,
    preprocess_data=pons_preprocess_data_args,
    training_configurations=[
        pons_train_model1_args,
        pons_train_model2_args,
    ],
    # training_resources=testing_training_resources,
    training_resources=default_training_resources,
)


larry_dataset_args = DownloadDatasetInterface(
    data_set_name="larry",
    # n_obs=500,
    # n_vars=2000,
)
larry_preprocess_data_args = PreprocessDataInterface(
    data_set_name=f"{larry_dataset_args.data_set_name}",
    adata=f"{larry_dataset_args.data_external_path}/{larry_dataset_args.data_set_name}.h5ad",
    vector_field_basis="emb",
)
larry_train_model1_args = PyroVelocityTrainInterface(
    adata=f"{larry_preprocess_data_args.data_processed_path}/{larry_dataset_args.data_set_name}_processed.h5ad",
    data_set_name=f"{larry_dataset_args.data_set_name}",
    model_identifier="model1",
    guide_type="auto_t0_constraint",
    # svi_train=True,
    batch_size=4000,
    offset=False,
    # cell_state="state_info",
    max_epochs=UNIFORM_MAX_EPOCHS,
)
larry_train_model2_args = PyroVelocityTrainInterface(
    adata=f"{larry_preprocess_data_args.data_processed_path}/{larry_dataset_args.data_set_name}_processed.h5ad",
    data_set_name=f"{larry_dataset_args.data_set_name}",
    model_identifier="model2",
    # svi_train=True,
    batch_size=4000,
    # cell_state="state_info",
    max_epochs=UNIFORM_MAX_EPOCHS,
)

larry_configuration = WorkflowConfiguration(
    download_dataset=larry_dataset_args,
    preprocess_data=larry_preprocess_data_args,
    training_configurations=[
        larry_train_model1_args,
        larry_train_model2_args,
    ],
    # training_resources=testing_training_resources,
    training_resources=large_training_resources,
)


if __name__ == "__main__":
    pass
